"""Configuration"""
import os.path
import tempfile

#: images server
IMAGES_ROOT = "https://images.krcg.org/"
#: what to download
VEKN_TWDA_URL = "http://www.vekn.fr/decks/twd.htm"
VEKN_VTES_URL = "http://www.vekn.net/images/stories/downloads/vtescsv_utf8.zip"
VEKN_VTES_LIBRARY_FILENAME = "vteslib.csv"
VEKN_VTES_CRYPT_FILENAME = "vtescrypt.csv"
VEKN_VTES_ENCODING = "utf_8_sig"
VEKN_VTES_I18N_URLS = {
    "fr-FR": (
        "http://www.vekn.net/images/stories/downloads/french/vtescsv_utf8.fr-FR.zip"
    ),
    "es-ES": (
        "http://www.vekn.net/images/stories/downloads/spanish/vtescsv_utf8.es-ES.zip"
    ),
}
VEKN_VTES_I18N_LIBRARY_FILENAME = "vteslib.%s.csv"
VEKN_VTES_I18N_CRYPT_FILENAME = "vtescrypt.%s.csv"
SUPPORTED_LANGUAGES = list(VEKN_VTES_I18N_URLS.keys())
#: where to store our data
TEMP_DIR = tempfile.gettempdir()
TWDA_FILE = os.path.join(TEMP_DIR, "krcg-twda.pkl")
VTES_FILE = os.path.join(TEMP_DIR, "krcg-vtes.pkl")
#: classic headers in deck lists
HEADERS = [
    "actin",
    "action",
    "acton",
    "allies",
    "ally",
    "average",
    "burn option",
    "capacity",
    "card",
    "combat",
    "comment",
    "conviction",
    "convinction",
    "crypt",
    "deck",
    "description",
    "disciplineless",
    "combo",
    "double",
    "equip",
    "equipament",
    "equipment",
    "equiptment",
    "equpment",
    "event",
    "librairie",
    "library",
    "master",
    "minion",
    "misc",
    "miscellaneous",
    "mixed",
    "mod",
    "modifier",
    "multitype",
    "non-skilled",
    "other",
    "politcal",
    "politic",
    "political",
    "power",
    "reaction",
    "rection",
    "retainer",
    "skill-less",
    "total",
    "trifle",
    "vote",
    "and",
]
#: official cards renaming
AKA = {
    # print change not registered in cards "Aka" field
    "mask of 1,000 faces": "Mask of a Thousand Faces",
}
#: widespread nicknames for clans
CLANS_AKA = {
    "!brujah": "Brujah antitribu",
    "!gangrel": "Gangrel antitribu",
    "!malkavian": "Malkavian antitribu",
    "!malk": "Malkavian antitribu",
    "!nosferatu": "Nosferatu antitribu",
    "!nosfe": "Nosferatu antitribu",
    "!salubri": "Salubri antitribu",
    "!salu": "Salubri antitribu",
    "!toreador": "Toreador antitribu",
    "!tore": "Toreador antitribu",
    "!tremere": "Tremere antitribu",
    "!trem": "Tremere antitribu",
    "!ventrue": "Ventrue antitribu",
    "brujah anti": "Brujah antitribu",
    "gangrel anti": "Gangrel antitribu",
    "malkavian anti": "Malkavian antitribu",
    "malk anti": "Malkavian antitribu",
    "nosferatu anti": "Nosferatu antitribu",
    "nosfe anti": "Nosferatu antitribu",
    "salubri anti": "Salubri antitribu",
    "salu anti": "Salubri antitribu",
    "toreador anti": "Toreador antitribu",
    "tore anti": "Toreador antitribu",
    "tremere anti": "Tremere antitribu",
    "trem anti": "Tremere antitribu",
    "ventrue anti": "Ventrue antitribu",
    "brother": "Blood Brother",
    "brothers": "Blood Brother",
    "giov": "Giovanni",
    "malk": "Malkavian",
    "naga": "Nagaraja",
    "nosfe": "Nosferatu",
    "salu": "Salubri",
    "tore": "Toreador",
    "trem": "Tremere",
    "cacophony": "Daughter of Cacophony",
    "daughter": "Daughter of Cacophony",
    "daughters": "Daughter of Cacophony",
    "set": "Follower of Set",
    "setite": "Follower of Set",
    "follower": "Follower of Set",
    "followers": "Follower of Set",
    "skulls": "Harbinger of Skulls",
    "harbinger": "Harbinger of Skulls",
    "harbingers": "Harbinger of Skulls",
    "trujah": "True Brujah",
}
#: disciplines nicknames/trigrams
DISC_AKA = {
    "abo": "Abombwe",
    "ani": "Animalism",
    "aus": "Auspex",
    "cel": "Celerity",
    "chi": "Chimerstry",
    "dai": "Daimoinon",
    "dem": "Dementation",
    "dom": "Dominate",
    "for": "Fortitude",
    "mal": "Maleficia",
    "mel": "Melpominee",
    "myt": "Mytherceria",
    "nec": "Necromancy",
    "obe": "Obeah",
    "obf": "Obfuscate",
    "obt": "Obtenebration",
    "pot": "Potence",
    "pre": "Presence",
    "pro": "Protean",
    "qui": "Quietus",
    "san": "Sanguinus",
    "ser": "Serpentis",
    "spi": "Spiritus",
    "str": "Striga",
    "tem": "Temporis",
    "tha": "Thaumaturgy",
    "than": "Thanatosis",
    "thau": "Thaumaturgy",
    "thn": "Thanatosis",
    "val": "Valeren",
    "vic": "Vicissitude",
    "vis": "Visceratika",
}
#: registered traits
TRAITS = [
    "primogen",
    "prince",
    "justicar",
    "inner circle",
    "imperator",
    "bishop",
    "archbishop",
    "cardinal",
    "regent",
    "priscus",
    "baron",
    "anarch",
    "laibon",
    "liaison",
    "magaji",
    "kholo",
    "black hand",
    "seraph",
    "red list",
    "infernal",
    "slave",
    "scarce",
    "sterile",
]
#: official disciplines trigrams used in crypt CSV
DIS_MAP = {
    "-none-": "none",
    "abo": "abombwe",
    "ani": "animalism",
    "aus": "auspex",
    "cel": "celerity",
    "chi": "chimerstry",
    "dai": "daimoinon",
    "def": "defense",
    "dem": "dementation",
    "dom": "dominate",
    "for": "fortitude",
    "inn": "innocence",
    "jud": "judgment",
    "mar": "martyrdom",
    "mel": "melpominee",
    "myt": "mytherceria",
    "nec": "necromancy",
    "obe": "obeah",
    "obf": "obfuscate",
    "obt": "obtenebration",
    "pot": "potence",
    "pre": "presence",
    "pro": "protean",
    "qui": "quietus",
    "red": "redemption",
    "san": "sanguinus",
    "ser": "serpentis",
    "spi": "spiritus",
    "tem": "temporis",
    "tha": "thaumaturgy",
    "thn": "thanatosis",
    "val": "valeren",
    "ven": "vengeance",
    "vic": "vicissitude",
    "vin": "vision",  # VIS is used in CSV
    "vis": "visceratika",
    "ABO": "ABOMBWE",
    "ANI": "ANIMALISM",
    "AUS": "AUSPEX",
    "CEL": "CELERITY",
    "CHI": "CHIMERSTRY",
    "DAI": "DAIMOINON",
    "DEM": "DEMENTATION",
    "DOM": "DOMINATE",
    "FOR": "FORTITUDE",
    "MEL": "MELPOMINEE",
    "MYT": "MYTHERCERIA",
    "NEC": "NECROMANCY",
    "OBE": "OBEAH",
    "OBF": "OBFUSCATE",
    "OBT": "OBTENEBRATION",
    "POT": "POTENCE",
    "PRE": "PRESENCE",
    "PRO": "PROTEAN",
    "QUI": "QUIETUS",
    "SAN": "SANGUINUS",
    "SER": "SERPENTIS",
    "SPI": "SPIRITUS",
    "TEM": "TEMPORIS",
    "THA": "THAUMATURGY",
    "THN": "THANATOSIS",
    "VAL": "VALEREN",
    "VIC": "VICISSITUDE",
    "VIS": "VISCERATIKA",
}
#: custom remap to match players abbreviations and typos
REMAP = {
    # our card parsing removes the dash
    "bang nakh": "Bang Nakh — Tiger's Claws",
    "khazar's diary": "Khazar's Diary (Endless Night)",
    "ohoyo hopoksia": "Ohoyo Hopoksia (Bastet)",
    # we not decode HTML properly as most of it is text
    "alia, god=92s messenger": "Alia, God's Messenger",
    "pentex=99 subversion": "Pentex™ Subversion",
    # traditions - players really do whatever they want
    "the first tradition": "First Tradition: The Masquerade",
    "first tradition": "First Tradition: The Masquerade",
    "1st tradition": "First Tradition: The Masquerade",
    "the second tradition": "Second Tradition: Domain",
    "second tradition": "Second Tradition: Domain",
    "2nd trad": "Second Tradition: Domain",
    "2nd tradition": "Second Tradition: Domain",
    "2nd tradition: domain": "Second Tradition: Domain",
    "the third tradition": "Third Tradition: Progeny",
    "third tradition": "Third Tradition: Progeny",
    "3rd tradition": "Third Tradition: Progeny",
    "the fourth tradition": "Fourth Tradition: The Accounting",
    "fourth tradition": "Fourth Tradition: The Accounting",
    "4th tradition": "Fourth Tradition: The Accounting",
    "4th tradition: accounting": "Fourth Tradition: The Accounting",
    "the fifth tradition": "Fifth Tradition: Hospitality",
    "fifth tradition": "Fifth Tradition: Hospitality",
    "5th tradition": "Fifth Tradition: Hospitality",
    "the sixth tradition": "Sixth Tradition: Destruction",
    "sixth tradition": "Sixth Tradition: Destruction",
    "6th tradition": "Sixth Tradition: Destruction",
    # hunting grounds
    "academic hg": "Academic Hunting Ground",
    "amusement park hg": "Amusement Park Hunting Ground",
    "asylum hg": "Asylum Hunting Ground",
    "base hg": "Base Hunting Ground",
    "campground hg": "Campground Hunting Ground",
    "corporate hg": "Corporate Hunting Ground",
    "fetish club hg": "Fetish Club Hunting Ground",
    "fetish hg": "Fetish Club Hunting Ground",
    "institution hg": "Institution Hunting Ground",
    "jungle hg": "Jungle Hunting Ground",
    "library hg": "Library Hunting Ground",
    "morgue hg": "Morgue Hunting Ground",
    "palace hg": "Palace Hunting Ground",
    "park hg": "Park Hunting Ground",
    "poacher's hg": "Poacher's Hunting Ground",
    "political hg": "Political Hunting Ground",
    "port hg": "Port Hunting Ground",
    "shanty Town hg": "Shanty Town Hunting Ground",
    "slum hg": "Slum Hunting Ground",
    "society hg": "Society Hunting Ground",
    "temple hg": "Temple Hunting Ground",
    "underworld hg": "Underworld Hunting Ground",
    "university hg": "University Hunting Ground",
    "uptown hg": "Uptown Hunting Ground",
    "warzone hg": "Warzone Hunting Ground",
    "zoo hg": "Zoo Hunting Ground",
    "academic h.g.": "Academic Hunting Ground",
    "amusement park h.g.": "Amusement Park Hunting Ground",
    "asylum h.g.": "Asylum Hunting Ground",
    "base h.g.": "Base Hunting Ground",
    "campground h.g.": "Campground Hunting Ground",
    "corporate h.g.": "Corporate Hunting Ground",
    "fetish club h.g.": "Fetish Club Hunting Ground",
    "institution h.g.": "Institution Hunting Ground",
    "jungle h.g.": "Jungle Hunting Ground",
    "library h.g.": "Library Hunting Ground",
    "morgue h.g.": "Morgue Hunting Ground",
    "palace h.g.": "Palace Hunting Ground",
    "park h.g.": "Park Hunting Ground",
    "poacher's h.g.": "Poacher's Hunting Ground",
    "political h.g.": "Political Hunting Ground",
    "port h.g.": "Port Hunting Ground",
    "shanty Town h.g.": "Shanty Town Hunting Ground",
    "slum h.g.": "Slum Hunting Ground",
    "society h.g.": "Society Hunting Ground",
    "temple h.g.": "Temple Hunting Ground",
    "underworld h.g.": "Underworld Hunting Ground",
    "university h.g.": "University Hunting Ground",
    "uptown h.g.": "Uptown Hunting Ground",
    "warzone h.g.": "Warzone Hunting Ground",
    "zoo h.g.": "Zoo Hunting Ground",
    "acad. hg": "Academic Hunting Ground",
    "univ. hg": "University Hunting Ground",
    "univ hg": "University Hunting Ground",
    # powerbases
    "pb: barranquilla": "Powerbase: Barranquilla",
    "pb: berlin": "Powerbase: Berlin",
    "pb: cape verde": "Powerbase: Cape Verde",
    "pb: chicago": "Powerbase: Chicago",
    "pb: los angeles": "Powerbase: Los Angeles",
    "pb: luanda": "Powerbase: Luanda",
    "pb: madrid": "Powerbase: Madrid",
    "pb:madrid": "Powerbase: Madrid",
    "pb: mexico city": "Powerbase: Mexico City",
    "pb: montreal": "Powerbase: Montreal",
    "pb:montreal": "Powerbase: Montreal",
    "pb: new york": "Powerbase: New York",
    "pb: rome": "Powerbase: Rome",
    "pb: savannah": "Powerbase: Savannah",
    "pb: tshwane": "Powerbase: Tshwane",
    "pb: washington, d.c.": "Powerbase: Washington, D.C.",
    "pb: zurich": "Powerbase: Zürich",
    "pb barranquilla": "Powerbase: Barranquilla",
    "pb berlin": "Powerbase: Berlin",
    "pb cape verde": "Powerbase: Cape Verde",
    "pb chicago": "Powerbase: Chicago",
    "pb los angeles": "Powerbase: Los Angeles",
    "pb luanda": "Powerbase: Luanda",
    "pb madrid": "Powerbase: Madrid",
    "pb mexico city": "Powerbase: Mexico City",
    "pb montreal": "Powerbase: Montreal",
    "pb new york": "Powerbase: New York",
    "pb rome": "Powerbase: Rome",
    "pb savannah": "Powerbase: Savannah",
    "pb tshwane": "Powerbase: Tshwane",
    "pb washington, d.c.": "Powerbase: Washington, D.C.",
    "pb zurich": "Powerbase: Zürich",
    "powerbase zurich": "Powerbase: Zürich",
    # punctuation
    "behind you": "Behind You!",
    "psyche": "Psyche!",
    # known abbreviations
    "antediluvian a.": "Antediluvian Awakening",
    "anthelios, the": "Anthelios, The Red Star",
    "archon inv.": "Archon Investigation",
    "call": "Call, The",
    "carlton": "Carlton Van Wyk",
    "con ag": "Conservative Agitation",
    "coven": "Coven, The",
    "direct": "Direct Intervention",
    "delaying": "Delaying Tactics",
    "dreams": "Dreams of the Sphinx",
    "dreams of the s.": "Dreams of the Sphinx",
    "effective": "Effective Management",
    "elysium": "Elysium: The Arboretum",
    "elysium: versailles": "Elysium: The Palace of Versailles",
    "entice": "Enticement",
    "felix fix": 'Felix "Fix" Hessian (Wraith)',
    "forced": "Forced Awakening",
    "forced aw": "Forced Awakening",
    "foreshadowing": "Foreshadowing Destruction",
    "golconda": "Golconda: Inner Peace",
    "govern": "Govern the Unaligned",
    "gtu": "Govern the Unaligned",
    "heidelberg": "Heidelberg Castle, Germany",
    "heidelburg": "Heidelberg Castle, Germany",
    "info highway": "Information Highway",
    "infohighway": "Information Highway",
    "info hwy": "Information Highway",
    "js simmons": "J. S. Simmons, Esq.",
    "js simmons esq": "J. S. Simmons, Esq.",
    "krc": "Kine Resources Contested",
    "krcg": "KRCG News Radio",
    "krcg news": "KRCG News Radio",
    "laptop": "Laptop Computer",
    "laptops": "Laptop Computer",
    "laptop comp": "Laptop Computer",
    "laptop comp.": "Laptop Computer",
    "legal manip": "Legal Manipulations",
    "london eve star": "London Evening Star, Tabloid Newspaper",
    "london evening star, tabloid": "London Evening Star, Tabloid Newspaper",
    "malk. dementia": "Malkavian Dementia",
    "masquer": "Masquer (Wraith)",
    "mister winthrop": "Mr. Winthrop",
    "molotov": "Molotov Cocktail",
    "owl": "Owl Companion",
    "patagia: flaps": "Patagia: Flaps Allowing Limited Flight",
    "pentex": "Pentex(TM) Subversion",
    "pto": "Protect Thine Own",
    "pulse": "Pulse of the Canaille",
    "rack": "Rack, The",
    "storage": "Storage Annex",
    "sudden": "Sudden Reversal",
    "talaq": "Talaq, The Immortal",
    "telepathic misdir.": "Telepathic Misdirection",
    "true love's kiss": "True Love's Face",
    "temptation of g.p.": "Temptation of Greater Power",
    "ventrue hq": "Ventrue Headquarters",
    "voter cap": "Voter Captivation",
    "wake with ef": "Wake with Evening's Freshness",
    "wake with e.f": "Wake with Evening's Freshness",
    "wake wef": "Wake with Evening's Freshness",
    "wwef": "Wake with Evening's Freshness",
    "wake...": "Wake with Evening's Freshness",
    "wake w/eve. freshness": "Wake with Evening's Freshness",
    "wake w/evening...": "Wake with Evening's Freshness",
    "wake": "Wake with Evening's Freshness",
    "wakes": "Wake with Evening's Freshness",
    "wakeys": "Wake with Evening's Freshness",
    "waste man op": "Waste Management Operation",
    "wmrh": "WMRH Talk Radio",
    "wwstick": "Weighted Walking Stick",
    # misspellings not fixed by difflib
    "2th tradition": "Second Tradition: Domain",
    "ancient influnse": "Ancient Influence",
    "blodd doll": "Blood Doll",
    "carver's meat packing plant": "Carver's Meat Packing and Storage",
    "carver's meat packing": "Carver's Meat Packing and Storage",
    "deflekion": "deflection",
    "denys": "Deny",
    "divine intervention": "Direct Intervention",
    "dogde": "Dodge",
    "dominat skill": "Dominate",
    "dominate: skillcard": "Dominate",
    "eagle sight": "Eagle's Sight",
    "sprit touch": "Spirit's Touch",
    "guard dog": "Guard Dogs",
    "golgonda": "Golconda: Inner Peace",
    "info superhighway": "Information Highway",
    "j.s. simons": "J. S. Simmons, Esq.",
    "judgement": "Judgment: Camarilla Segregation",
    "krcg newspaper": "KRCG News Radio",
    "krcg radio station": "KRCG News Radio",
    "lost in the crowd": "Lost in Crowds",
    "lost n the crowds": "Lost in Crowds",
    "milicent smith: vampire hunter": "Millicent Smith, Puritan Vampire Hunter",
    "obfuscate skill": "Obfuscate",
    "ps: istanbul": "Praxis Seizure: Istanbul",
    "rejuvenation": "Rejuvenate",
    "rumour mill tabloid": "The Rumor Mill, Tabloid Newspaper",
    "rumor mill, the": "The Rumor Mill, Tabloid Newspaper",
    "soul gems": "Soul Gem of Etrius",
    "tomb of rameses the cheesemonger": "Tomb of Rameses III",
    "truth of a 1000 lies": "Truth of a Thousand Lies",
    "veil of legions": "Veil the Legions",
    # 🥚
    "enkil cock": "enkil cog",
    "parity shit": "Parity Shift",
    "heart of the shitty": "Heart of the City",
    "heart of cheating": "Heart of Nizchetus",
    "hero fellation": "Aire of Elation",
}
#: type order for deck display
TYPE_ORDER = [
    "Master",
    "Conviction",
    "Action",
    "Action/Combat",
    "Action/Reaction",
    "Ally",
    "Equipment",
    "Political Action",
    "Retainer",
    "Power",
    "Action Modifier",
    "Action Modifier/Combat",
    "Action Modifier/Reaction",
    "Reaction",
    "Combat",
    "Combat/Reaction",
    "Event",
]
#: traits are hard to parse: manual fixups
SEARCH_EXCEPTIONS = {
    "The Baron": ["-trait:anarch"],
    "Gwen Brand": [
        "+discipline:ANIMALISM",
        "+discipline:AUSPEX",
        "+discipline:CHIMERSTRY",
        "+discipline:FORTITUDE",
    ],
}
#: some decks in the TWDA do not respect the rules when it comes to deck size
TWDA_CHECK_DECK_FAILS = {
    "2k9avangarda",  # 59 cards listed
    "2k8glqmich",  # 91 cards listed
    "2k8sanfranqual",  # 59 cards listed
    "2k8pwbsla2",  # 91 cards listed
    "2k6faceaface",  # 91 cards listed
    "2k4virolaxboston",  # 91 cards listed
    "2k4edith",  # 91 cards listed
    "2k4pariscup",  # 11 crypt cards listed
    "2k3nycanarch",  # 91 cards listed
    "saveface2k1",  # 91 cards listed
    "genconuk2k1-treasure",  # 91 cards listed
    "jd32000",  # 91 cards listed
    "dog",  # 100 cards listed
    "matt-alamut",  # 91 cards listed
    "stevewampler",  # 59 cards listed
}
# Legacy, kept here just in case
# fixes done to problematic decklists in the TWDA.
# KRCG was used to regenerate full proper decklists in the twda_fix folder
# TWDA_FIXUP = {
#     # ill-formatted comments
#     "2010hunrb": {"cards_comments": {"Deep Song": "just the right number"}},
#     "2010pwblaQ": {
#         "cards_comments": {
#             "The Art of Memory": (
#                 "Linchpin card for dialing in the moment-appropriate card."
#             ),
#             "Touch of Clarity": ("Recent addition, unneeded in this tournament."),
#         }
#     },
#     "2010pwbla1": {
#         "cards_comments": {
#             "Ancient Influence": (
#                 'eradicate your prey and call as a "spare" Tupdog action'
#             ),
#             "Reins of Power": (
#                 'eradicate your predator and call as a "spare" Tupdog action'
#             ),
#         }
#     },
#     "2k9mojlp": {
#         "cards": {  # 4 cards listed twice
#             "Conditioning": 6,
#             "Leverage": 3,
#             "Seduction": 4,
#             "The Sleeping Mind": 5,
#         }
#     },
#     "2k8sequeenslandcq": {
#         "cards": {
#             "Praxis Seizure: Geneva": 4,  # all praxis cards on a single line
#             "Praxis Seizure: Berlin": 4,
#             "Praxis Seizure: Cairo": 2,
#             "Elder Impersonation": 3,  # wrong comment mark
#         },
#         "cards_comments": {
#             "Elder Impersonation": "would exchange for Faceless Night x4"
#         },
#     },
#     "2k8torunminiq": {
#         "cards": {
#             "Direct Intervention": 2,
#             "Powerbase: Tshwane": 1,
#             "WMRH Talk Radio": 1,
#             "No Secrets From the Magaji": 3,
#             "Predator's Transformation": 5,
#             "Ancilla Empowerment": 1,
#             "Superior Mettle": 2,
#             "Predator's Communion": 4,
#         },
#         "cards_comments": {
#             "Direct Intervention": (
#                 "I'm thinking about adding another one, this "
#                 "card is cruicial for any block denial actions, or things that "
#                 "simply screw your spiders"
#             ),
#             "Powerbase: Tshwane": (
#                 "there was some magic around this card, I got it "
#                 "in every single game, in the first 10 cards drawn"
#             ),
#             "WMRH Talk Radio": (
#                 "didn't use it at all, metagame wasn't stealth heavy though"
#             ),
#             "No Secrets From the Magaji": (
#                 "was thinking about 4-5, but 3 is a good "
#                 "number, you draw too much attention if you put it on too early"
#             ),
#             "Predator's Transformation": (
#                 "wonderful card, can play it anytime, and "
#                 "the bonus superior effect did help a lot many times"
#             ),
#             "Ancilla Empowerment": (
#                 "nobody really expects these 2 until its too late"
#             ),
#             "Superior Mettle": (
#                 "I'd revise the prevent cards anyway, maybe adding "
#                 "some with no cost, or more of SM. it worked nice in overall though"
#             ),
#             "Predator's Communion": (
#                 "very helpful before you get no secrets, easy " "to drop afterwards"
#             ),
#         },
#     },
#     "2k7campeonatojuizforano": {
#         "cards_comments": {
#             "Direct Intervention": (
#                 "saved me a lot of times, unfortunately I couldn't pack more than "
#                 "one."
#             ),
#             "The Slaughterhouse": (
#                 "useful either to speed deck depletion or to trade for something "
#                 "useful under Anthelios."
#             ),
#             "Wash": (
#                 "not as effective as I expected, but also not a hassle because it's "
#                 "trifle."
#             ),
#         }
#     },
#     "2k7comjust": {
#         "cards_comments": {
#             "Abbot": (
#                 "I like these. Got smiled at until people realized I actually packed "
#                 "combat."
#             ),
#             "Gramle": (
#                 "This should help contesting Aranthebes, fetch another AI, or DI for "
#                 "a delayed vote. Never used though."
#             ),
#             "Botched Move": (
#                 "Pushes the maximum damage in a round to 7. Cute, but nearly "
#                 "useless. :)"
#             ),
#         }
#     },
#     "2k7woaboston": {
#         "cards": {
#             "Perfectionist": 2,
#             "Moise Kasavubu": 1,
#             "Vagabond Mystic": 1,
#             "The Crusader Sword": 1,
#             "Determine": 2,
#         },
#         "cards_comments": {
#             "Perfectionist": (
#                 "weren't too useful - the Vamps didn't act too often.  Maybe -1, +1 "
#                 "Blood Doll"
#             ),
#             "Moise Kasavubu": (
#                 "seems good, but never got him out to confirm - in past uses, he's "
#                 "just gotten blocked, but with the stealth I put him back in"
#             ),
#             "Vagabond Mystic": (
#                 "undecided; life was rarely an issue, except when a little healing "
#                 "wouldn't have mattered"
#             ),
#             "The Crusader Sword": (
#                 "in retrospect, IR Goggles to go with this might be nice"
#             ),
#             "Determine": (
#                 "never used; took too much vsn out of the crypt, and rarely needed"
#             ),
#         },
#     },
#     "2k4volkerpit": {
#         "comments": (
#             "Description: Lasombra S&B for the tournament in Pittsburgh, 3/27/04\n\n"
#             "Anyway, here's the decklist. There's not much new tech in here but for "
#             "some\n"
#             "concerted effort to keep blood on my vampires. I've found that the "
#             "Hungry\n"
#            "Coyote is close to miraculous for keeping alive if you feel you have to\n"
#            "turtle up for a little while,and it's usually a turn well spent hunting\n"
#             "with the HC around to recoup some blood spent on Dominate bleeds.\n\n"
#             "Possible improvements:\n\n"
#             "Drop one Kassiym from an extra copy of Banjoko, add an extra Path\n"
#             "and reduce the number of Stone Travels for some extra combat defence.\n"
#             "Maybe drop the prayer cards (Dramatic Upheaval, Graverobbing) and try\n"
#             "to find some more Shroud of Absence... As it is, though, the deck\n"
#             "performs well and cycles beautifully. True combat as a predator will\n"
#           "likely crunch it, but true combat has to go upstream heavily or it won't\n"
#             "survive, so you may be able to just sit it out for a while until\n"
#             "the coast is clear.\n"
#         )
#     },
#     "2k3ec": {
#         "cards": {"Smiling Jack, The Anarch": 3, "Auspex": 2},
#         "cards_comments": {
#             "Smiling Jack, The Anarch": "to many!",
#             "Auspex": "don't need them, maybe one!",
#         },
#     },
#     "2k7palmaqual": {"cards": {"Animalism": 1}},
#     "2k6aussiechamp": {"cards": {"Dominate": 1}},
#     "2k6sonarba": {"cards": {"Dominate": 4}},
#     "2k6praxispoitiers": {"cards": {"Dominate Kine": 2}},
#     "2k6faceaface": {"cards": {"The Coven": 1}},
#     "2k6nerq-templecon": {"cards": {"Dominate": 3}},
#     "2k5leicesteraug": {"cards": {"Presence": 2}},
#     "2k5valhallajuly": {"cards": {"Wake with Evening's Freshness": 5}},
#     "2k5ropecon": {"cards": {"Dominate": 1}},
#     "2k5blackutrecht": {"cards": {"Fear of Mekhet": 1, "Fortune Teller Shop": 1}},
#     "2k5dreamboston": {"cards": {"Dominate": 1}},
#     "2k4norms": {"cards": {"Auspex": 1, "Obfuscate": 1}},
#     "2k4littlesister": {"cards": {"Celerity": 2}},
#     "2k4gks": {"cards": {"Dominate": 1}},
#     "2k3italyqualifier": {
#         "cards": {
#             "Jake Washington": 3,
#             "The Hungry Coyote": 1,
#             "The Embrace": 3,
#             "Entrancement": 3,
#             "The Summoning": 5,
#             "Change of Target": 6,
#             "Marijava Ghoul": 2,
#         },
#         "cards_comments": {
#             "Jake Washington": (
#                 "He is perfect for both the WG and let my midcaps eat if hungry"
#             ),
#             "The Hungry Coyote": (
#                 "after the tournament I have decided that the Jakes are enough for "
#               "this and that my minions have always better actions to do rather than "
#                 'hunt. I would change it for a Pentex "anti helena/Lucas halton/..." '
#                 "Subversion"
#             ),
#             "The Embrace": (
#                 "extremely useful for bleed and multiple action as rescuing from "
#                 "torpor/eating alive/rush walls"
#             ),
#             "Entrancement": (
#                 "To defend my omelettes, for bleeding or for an extrarecruitment if "
#                 "ever able, this card is so great"
#             ),
#             "The Summoning": (
#                 "key, essential, If you summon a WG and the action is blocked you "
#                 "don't lose the Meat machine so you could try it later, and it also "
#                 "multiplies the number of getting a WG from my library"
#             ),
#             "Change of Target": (
#               "for the embraces aswell and the Ghouls if an evil killer vamps untaps "
#                 "suddenly while attacking"
#             ),
#             "Marijava Ghoul": (
#                "exceptionals, There are 15 pressence action cards and this makes 'em "
#                 "more efficient, but he is also meat for the WG god"
#             ),
#         },
#     },
#     "2k3frenchecq": {"cards": {"Obfuscate": 1}},
#     "2k3helsinki5": {"cards": {"Wake with Evening's Freshness": 5}},
#     "2k3sdgothenburg": {"cards": {"Obfuscate": 3}},
#     "2k3cadaverous": {"cards": {"Falcon's Eye": 4}},
#     "2k3winnipegapril": {
#         "cards": {"Bum's Rush": 4},
#         "cards_comments": {
#             "Bum's Rush": "Need some Harass in here for 2nd round fun",
#             "Conditioning": (
#                 "used only at lesser to avoid archons. Only 1 vamp had superior dom."
#             ),
#         },
#     },
#     "2k3rome": {"cards": {"Presence": 7}},
#     "2k3psmilwaukee": {
#         "comments": (
#             "MOTTO: Coming soon to a school board near you.\n\n"
#             "Notes:\n\n"
#             "DO NOT SCARE THE TABLE... if at all convenient.\n\n"
#             "Get out Arika... burn people for fun and profit.\n"
#             "Profit is Minion Tap/Voter Cap. Use the Governs\n"
#             "to bring out friends for Arika. Relies on Zillah's\n"
#             "to outrun Combat Decks to a vampire, who can then be\n"
#             "burned with PTO, and fails utterly to weenie combat,\n"
#             "fast big combat, Camarilla combat, and any of the same\n"
#             "if it's bleed.\n\n"
#             "This deck came into being preNERFing of 5th Tradition.\n"
#             "I put it together to see if I could still get big\n"
#             "vamp bloat to work. I managed to guess the local\n"
#             "metagame to a T on the sect issues required to play\n"
#             "the deck. Only one deck I saw was all Camarilla.\n\n"
#             "Changes: Some GtU to Mind Rape, 3 Lucinde 2 Vitel\n"
#             "lots more combat D/bleed D depending on metagame.\n"
#             "I've added all the Awe's I own, more may be good.\n\n"
#             "Cameron Domer\n"
#             "orcaorcinus@hotmail.com\n"
#         )
#     },
#     "2k3gothenburgblizzard": {"cards": {"Celerity": 3}},
#     "2k2eclastq": {"cards": {"Jake Washington": 5}},
#     "2k2sydney": {
#         "cards": {"Majesty": 2},
#         "cards_comments": {
#           "Majesty": "after CE probably swap out for dodge, ecstasy, change of target"
#         },
#     },
#     # guessing "Manouever" was "Fake Out"
#     "2k2ukqualifier": {"cards": {"Fake Out": 4}},
#     "2k2watfordmay": {
#         "author": (
#             "Tony 'literary Thug' Smith, with a bit of creative inspiration from "
#             "Jon Cooper."
#         ),
#         "comments": (
#             "Didn't have too many problems on the day and the combination of bone\n"
#             "crunching combat and 2nd tradition seemed to handle most things. A\n"
#             "little slow getting the first point in the final, which nearly\n"
#             "scuppered my plans.\n\n"
#             "The first version of this had 4 Don Cruez. I later decided that using\n"
#             "a smaller crypt and another Protean master was probably the way to go.\n"
#             "I may even drop the Dons to 2 and add a second Beast. I usually went\n"
#             "for either Theo or Volker first, on the basis that the Don was just\n"
#             "too big. Although, to be fair, his maneuver did save me a couple of\n"
#             "times against the Wookie's assamites.\n\n"
#             "Not much in the way of subtlety, but hey, burning vampires is fun.\n\n"
#             "Remember kids, don't do drugs.\n\n"
#             "Tony\n\n"
#             "Literary Thug\n"
#         ),
#     },
#     # replaced 6x "Praxis Seizure..." with Jyhad non-clan generic Praxis
#     "2k2pariseq": {
#         "cards": {
#             "Presence": 3,
#             "Praxis Seizure: Atlanta": 1,
#             "Praxis Seizure: Boston": 1,
#             "Praxis Seizure: Chicago": 1,
#             "Praxis Seizure: Cleveland": 1,
#             "Praxis Seizure: Dallas": 1,
#             "Praxis Seizure: Houston": 1,
#         }
#     },
#     "2k2gothenburgapr": {"cards": {"Presence": 10}},
#     "2k2during": {"cards": {"Fortitude": 5}},
#     "2k2newyork": {
#         "cards": {
#             "Catatonic Fear": 14,
#             "Marijava Ghoul": 1,
#         },
#         "cards_comments": {
#             "Catatonic Fear": "and i needed them all+majesty",
#             "Marijava Ghoul": "needed to break free-+1 stealth :)",
#         },
#     },
#     "2k2orccon2": {"cards": {"Presence": 1}},
#     # because of AK-47 we can't parse dash (-) post counts with no space
#     "2k2torrancejan": {
#         "cards": {
#             "Antediluvian Awakening": 1,
#             "Information Highway": 2,
#             "The Parthenon": 2,
#             "Minion Tap": 2,
#             "Blood Doll": 4,
#             "Sudden Reversal": 8,
#             "Kindred Spirits": 18,
#             "Eyes of Chaos": 10,
#             "Confusion": 6,
#             "Swallowed by the Night": 8,
#             "Cloak the Gathering": 6,
#             "Elder Impersonation": 4,
#             "Spying Mission": 7,
#             "Faceless Night": 6,
#             "Lost in Crowds": 6,
#         }
#     },
#     "2k2gothenburg": {"cards": {"Dominate": 1, "Thaumaturgy": 1}},
#     "2k3tcithaca": {
#         # replace "praxis seizure (various)" with Jyhad non-clan-specific ones
#         "cards": {
#             "Praxis Seizure: Atlanta": 1,
#             "Praxis Seizure: Boston": 1,
#             "Praxis Seizure: Chicago": 1,
#             "Praxis Seizure: Cleveland": 1,
#         },
#     },
#     "gothenburg2k1": {"cards": {"Auspex": 4}},
#     "watford-ts": {
#         "cards": {
#             "Hostile Takeover": 2,
#             "Wake with Evening's Freshness": 5,
#             "Ivory Bow": 1,
#             "Seduction": 10,
#         },
#         "cards_comments": {
#             "Hostile Takeover": (
#                 "I know that there are only 2 ventru in the crypt but trust me, the "
#                 "hostiles have won games for me more than once"
#             ),
#             "Wake with Evening's Freshness": "changed from 4 at the last minute",
#             "Ivory Bow": "didn't see it all day, may need to come out",
#             "Seduction": (
#                 "the numbers of both of these has fluctuated between 8 and 12. "
#                 "for this tourney I went for the happy medium."
#             ),
#         },
#     },
#     # misplaced card name in tail comment got included in the deck
#     "darbyla2k1": {"cards": {"Social Charm": 15}},
#     "joshphiladelphia": {"cards": {"Protean": 1}},
#     "damnans": {"cards": {"Fortitude": 2}},
#     # card type prefix - too rare to bother parsing this
#     "ckgc2k": {
#         "comments": "",
#         "cards": {
#             "Govern the Unaligned": 6,
#             "Bum's Rush": 1,
#             "Bonding": 7,
#             "Conditioning": 6,
#             "Shadow Play": 8,
#             "Shroud of Night": 3,
#             "Arms of the Abyss": 3,
#             "Entombment": 3,
#             "Shadow Body": 5,
#             "Leather Jacket": 4,
#             "Archon Investigation": 1,
#             "Blood Doll": 4,
#             "Minion Tap": 2,
#             "Dominate": 3,
#             "Obtenebration": 2,
#             "Elysian Fields": 3,
#             "Elysium: The Arboretum": 2,
#             "Political Hunting Ground": 3,
#             "Slave Auction": 2,
#             "Sudden Reversal": 1,
#             "Deflection": 5,
#             "Eyes of the Night": 5,
#             "Redirection": 6,
#             "Wake with Evening's Freshness": 6,
#         },
#     },
#     # because of AK-47 we can't parse dash (-) post counts with no space
#     "swkc": {
#         "comments": "",
#         "cards": {
#             "Auspex": 2,
#             "Presence": 4,
#             "Blood Doll": 2,
#             "Ascendance": 2,
#             "Misdirection": 2,
#             "Sudden Reversal": 2,
#             "Short-Term Investment": 1,
#             "Minion Tap": 1,
#             "Major Boon": 1,
#             "Aching Beauty": 1,
#             "Society Hunting Ground": 1,
#             "Art Museum": 1,
#             "Forced Awakening": 2,
#             "Marijava Ghoul": 2,
#             "Palatial Estate": 1,
#             "Telepathic Counter": 9,
#             "Telepathic Misdirection": 11,
#             "Kindred Intelligence": 2,
#             "Spirit's Touch": 2,
#             "Precognition": 2,
#             "Eagle's Sight": 1,
#             "Enhanced Senses": 1,
#             "Pulse of the Canaille": 1,
#             "Majesty": 12,
#             "Enchant Kindred": 8,
#             "Legal Manipulations": 4,
#             "Entrancement": 1,
#             "Aire of Elation": 2,
#             "Propaganda": 1,
#             "Media Influence": 2,
#             "Staredown": 2,
#             "Catatonic Fear": 1,
#             "Social Charm": 3,
#         },
#     },
#     "quebec": {"cards": {"Fortitude": 1, "Presence": 1}},
#     "leedec99": {"cards": {"Obfuscate": 1}},
#     "portoct99": {"cards": {"Celerity": 1}},
#     "matt-alamut": {
#         "cards": {
#             "Obfuscate": 1,
#             "Presence": 1,
#             "Praxis Seizure: Istanbul": 1,
#             "Praxis Seizure: Paris": 1,
#             "Praxis Seizure: Monaco": 1,
#             "Praxis Seizure: Dublin": 1,
#         }
#     },
#     "Brian": {"cards": {"Dominate": 2, "Presence": 1}},
#     "rob": {"cards": {"Dominate": 1, "Fortitude": 1}},
# }



"""Static files generator for third parties.

Produces static files for use in third parties softwares.
"""

import argparse
import collections
import hashlib
import json
import os.path
import re
import sys

import requests

from . import logging
from . import twda
from . import utils
from . import vtes

logger = logging.logger

parser = argparse.ArgumentParser(prog="krcg", description="VTES tool")


class FormatChoice(utils.NargsChoiceWithAliases):
    CHOICES = ["standard", "amaranth"]


parser.add_argument(
    "formats",
    metavar="FORMAT",
    nargs="+",
    action=FormatChoice,
    help="Generate static files in given formats",
)


def standard_rulings() -> dict:
    """Rulings in JSON format."""
    result = {
        "rulings": {},
        "references": {},
        "cards": collections.defaultdict(list),
    }
    for key, data in vtes.VTES.items():
        if not isinstance(key, int):
            continue
        ruling_data = vtes.VTES.rulings.get(vtes.VTES.get_name(data))
        if not ruling_data:
            continue
        for ruling in ruling_data.get("Rulings", []):
            h = hashlib.sha3_256(ruling.encode("utf-8")).hexdigest()[-12:]
            result["rulings"].setdefault(h, ruling)
            result["cards"][key].append(h)
        for link in ruling_data.get("Rulings Links", []):
            result["references"].setdefault("[" + link["Reference"] + "]", link["URL"])
    return result


def amaranth_twda() -> list:
    """TWDA for Amaranth."""
    try:
        amaranth_ids = {
            vtes.VTES.get_name(card["name"]): card["id"]
            for card in (
                requests.get("http://amaranth.vtes.co.nz/api/cards").json()["result"]
            )
            if card["name"] in vtes.VTES
        }
    except requests.exceptions.ConnectionError as e:
        logger.exception("failed to connect to {}", e.request.url)
        exit(1)
    ret = []
    for twda_id, deck in twda.TWDA.items():
        try:
            ret.append(
                {
                    "author": deck.player,
                    "date": deck.date.isoformat(),
                    "title": deck.name or "",
                    "twda_key": twda_id,
                    "description": deck.event + ", " + deck.place,
                    "cards": {amaranth_ids[k]: v for k, v in deck.items()},
                }
            )
        except KeyError as e:
            logger.error(
                "TWDA: {card} not referenced in Amaranth, skipping {twda_id}",
                card=e.args[0],
                twda_id=twda_id,
            )
            continue
    return ret


def standard_twda() -> str:
    """A normalized HTML version of the TWDA"""
    content = """<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">
<html><head><meta http-equiv="Content-Type" content="text/html; charset=utf-8">
<title>TWDA</title></head>

<body>
<center>
<h1>Tournament Winning Deck Archive</h1>
<h2>Formerly hosted on the Lasombra</h2>

<p>
This is a chronological archive of decks that have won tournaments that were<br>
sanctioned by the Vampire: Elder Kindred Network and had 10 or more players.<br>
National and Continental Championship are included as well, irrelatively to their attendance.<br>
To be included in this list, a tournament report had to be made on the official V:EKN Player's Forum:<br>
<a href="http://www.vekn.net/index.php/forum/9-event-reports-and-twd">http://www.vekn.net/index.php/forum/9-event-reports-and-twd</a><br>
This archive will not include the Storyline Tournament Winning Decks.<br>
Unless otherwise noted, each deck was the winning player's creation.<br>
<br>
The original <a href="http://www.thelasombra.com/hall_of_fame.htm">Hall of Fame</a> and the <a href="http://www.veknfrance.com/decks/twd.htm">Tournament Winning Deck Archive</a> were closed as of October 26, 2013.<br>
Many thanks to Jeff Thompson for maintaining them for all these years.
</p>
<table width="500" align="center" border="2" cellpadding="0">
<tbody><tr align="center">
<td colspan="10">You can jump to the end of a year that interests you:</td>
</tr>
<tr align="center">
<td colspan="4"></td>
<td><a href="#Year2020">2020</a></td>
<td><a href="#Year2019">2019</a></td>
<td><a href="#Year2018">2018</a></td>
<td><a href="#Year2017">2017</a></td>
<td colspan="3"></td>
</tr><tr align="center">
<td><a href="#Year2016">2016</a></td>
<td><a href="#Year2015">2015</a></td>
<td><a href="#Year2014">2014</a></td>
<td><a href="#Year2013">2013</a></td>
<td><a href="#Year2012">2012</a></td>
<td><a href="#Year2011">2011</a></td>
<td><a href="#Year2010">2010</a></td>
<td><a href="#Year2009">2009</a></td>
<td><a href="#Year2008">2008</a></td>
<td><a href="#Year2007">2007</a></td>
</tr><tr align="center">
<td><a href="#Year2006">2006</a></td>
<td><a href="#Year2005">2005</a></td>
<td><a href="#Year2004">2004</a></td>
<td><a href="#Year2003">2003</a></td>
<td><a href="#Year2002">2002</a></td>
<td><a href="#Year2001">2001</a></td>
<td><a href="#Year2000">2000</a></td>
<td><a href="#Year1999">1999</a></td>
<td><a href="#Year1998">1998</a></td>
<td><a href="#Year1997">1997</a></td>
</tr>
</tbody></table>
<p>Some abbreviations are used throughout the archive:
<table>
<tr><td>NC</td><td>National Championship</td><td>NAC</td><td>North American (Continental) Championship</td></tr>
<tr><td>NCQ</td><td>National Championship Qualifier</td><td>SAC</td><td>South American (Continental) Championship</td></tr>
<tr><td>ECQ</td><td>European Championship Qualifier</td><td>EC</td><td>European (Continental) Championship</td></tr>
<tr><td>CCQ</td><td>Continental Championship Qualifier</td><td>ACC</td><td>Asian Continental Championship</td></tr>
</table>"""  # noqa
    decks = sorted(twda.TWDA.items(), key=lambda a: a[1].date, reverse=True)
    current_year = None
    for twda_id, deck in decks:
        if current_year != deck.date.year:
            current_year = deck.date.year
            content += f'\n<h3><a id="Year{current_year}">{current_year}</a></h3>\n\n'
        player = deck.player
        assert player, f"no player indicated for deck #{twda_id}"
        if player[-1] != "s":
            player += "'s"
        else:
            player += "'"
        event = deck.event
        event = re.sub(r"\s*--\s+.*", "", event)
        place = deck.place
        place = re.sub(r"\s*,", ",", place)
        place = re.sub(r",,", ",", place)
        place = [x.strip() for x in place.split(",")[-2:]]
        event = event.strip().strip(":")
        if event[-len(place[0]) :] == place[0]:
            place = place[1:]
            event = event.strip().strip(":") + ","
        else:
            event = event.strip().strip(":") + ":"
        place = ", ".join(place)
        content += f"<a href=#{twda_id}>{player} {event} {place} {deck.date.strftime('%B %Y')}</a><br>\n"  # noqa: E501
    content += "</center>\n"
    for twda_id, deck in decks:
        content += f"<a id={twda_id} href=#>Top</a>\n<hr><pre>\n"
        content += deck.to_txt(wrap=False)
        content += "\n</pre>\n"
    return content


def main():
    """Entrypoint for the krcg-gen tool."""
    args = parser.parse_args(sys.argv[1:])
    try:
        if not vtes.VTES:
            vtes.VTES.load_from_vekn(save=False)
        vtes.VTES.configure()
        if not twda.TWDA:
            twda.TWDA.load_from_vekn(save=False)
        twda.TWDA.configure(spoilers=False)
    except requests.exceptions.ConnectionError as e:
        logger.exception("failed to connect to {}", e.request.url)
        exit(1)
    out_folder = os.path.join(os.path.dirname(__file__), "..", "static")
    if "standard" in args.formats:
        with open(
            os.path.join(out_folder, "standard-rulings.json"),
            "w",
            encoding="utf-8",
        ) as out_file:
            json.dump(standard_rulings(), out_file, indent=2)
        with open(
            os.path.join(out_folder, "twd.htm"),
            "w",
            encoding="utf-8-sig",
        ) as out_file:
            out_file.write(standard_twda())
    if "amaranth" in args.formats:
        with open(
            os.path.join(out_folder, "amaranth-twda.json"),
            "w",
            encoding="utf-8",
        ) as out_file:
            json.dump(amaranth_twda(), out_file, indent=2)



"""Deck class: pickling and card access under conditions"""
from typing import Callable, Dict, Generator, List, Optional, TextIO, Tuple
import collections
import datetime
import itertools
import re
import textwrap

import arrow
import arrow.parser

from . import config
from . import logging
from . import parser
from . import vtes
from . import utils


Condition = Optional[Callable[[str], bool]]
CardsCount = Tuple[str, int]
CardsList = List[Dict[str, int]]
CardsBlock = Tuple[str, CardsList]

logger = logging.logger


class Deck(collections.Counter):
    """A VTES deck, including meta information such as author and decription.

    Attributes:
        - id (str): ID of the deck. Official TWDA id for TWDA decks
        - name (str): Name of the deck
        - author (str): Deck author
        - comments (str): Comments on the deck
        - event (str): Event (for TWD)
        - place (str): Place where the event was held
        - date (datetime.date): Date on which the event was held
        - tournament_format (str): Format of the event
        - players_count (int): Count of players at the event
        - player (str): Player who played the deck
        - event_link (str): Link to the event webpage
        - score (str): Score the deck achieved (XgwY - Zvp in the final)
    """

    def __init__(self, *args, **kwargs):
        self.id = kwargs.pop("id", None)
        self.author = kwargs.pop("author", None)
        super().__init__(*args, **kwargs)
        self.event = None
        self.place = None
        self.date = None
        self.tournament_format = None
        self.players_count = 0
        self.player = None
        self.event_link = None
        self.score = None
        self.name = None
        self.cards_comments = {}
        self.comments = ""
        # very corner case: counting the number of "Raven" in "Camille Devereux" copies
        self.raven = 0
        self.parser = parser.Parser(self)

    @classmethod
    def from_txt(
        cls,
        input: TextIO,
        id: str = None,
        author: str = None,
        offset: int = 0,
        twda: bool = False,
    ):
        p = parser.Parser(cls(id=id, author=author))
        p.parse(input, offset, twda)
        return p.deck

    def check(self) -> bool:
        """Check a deck conforms to the rules IRT cards count

        Does not check grouping nor banished cards

        Returns:
            True if the deck checks out
        """
        res = True
        library_count = self.cards_count(vtes.VTES.is_library)
        crypt_count = self.cards_count(vtes.VTES.is_crypt)
        if library_count < 60:
            logger.warning(
                "deck has too few cards ({count}) [{repr}]",
                id=self.id,
                count=library_count,
                repr=self,
            )
            res = False
        if library_count > 90:
            logger.warning(
                "deck has too many cards ({count}) [{repr}]",
                id=self.id,
                count=library_count,
                repr=self,
            )
            res = False
        if crypt_count < 12:
            logger.warning(
                "deck is missing crypt cards ({count}) [{repr}]",
                id=self.id,
                count=crypt_count,
                repr=self,
            )
            res = False
        return res

    def __getstate__(self):
        """For pickle serialization.

        Deck inherits `dict` and its special handling of pickle.
        """
        return {
            "id": self.id,
            "cards": collections.OrderedDict(self.cards()),
            "author": self.author,
            "event": self.event,
            "place": self.place,
            "date": self.date.isoformat() if self.date else None,
            "tournament_format": self.tournament_format,
            "score": self.score,
            "players_count": self.players_count,
            "player": self.player,
            "event_link": self.event_link,
            "name": self.name,
            "cards_comments": self.cards_comments,
            "comments": self.comments,
            "raven": self.raven,
        }

    def __setstate__(self, state):
        """For pickle deserialization."""
        self.id = state.get("id")
        self.author = state.get("author")
        self.event = state.get("event")
        self.place = state.get("place")
        try:
            self.date = datetime.date.fromisoformat(state.get("date") or "")
        except ValueError:
            pass
        self.tournament_format = state.get("tournament_format")
        self.score = state.get("score")
        self.players_count = int(state.get("players_count", 0))
        self.player = state.get("player")
        self.event_link = state.get("event_link")
        self.name = state.get("name")
        self.comments = state.get("comments", "")
        self.cards_comments = state.get("cards_comments", {})
        self.raven = state.get("raven", {})
        self.update(state.get("cards", {}))

    def __reduce__(self):
        """For pickle serialization."""
        return (Deck, (), self.__getstate__())

    def __str__(self):
        return self.name or "(No Name)"

    def cards(self, condition: Condition = None) -> Generator[CardsCount, None, None]:
        """Generator yielding (card_name, count), with an optional filter.

        Args:
            condition: Condition each card must validate to be selected

        Yields:
            card_name, count
        """
        for card, count in self.items():
            if condition and not condition(card):
                continue
            yield card, count

    def card_names(self, condition: Condition = None) -> Generator[str, None, None]:
        """Generator yielding card names with an optional filter.

        Args:
            condition: Condition each card must validate to be selected

        Yields:
            card_name
        """
        for card, _count in self.cards(condition):
            yield card

    def cards_count(self, condition: Condition = None) -> int:
        """Card counts with an optional filter.

        Args:
            condition: Condition each card must validate to be selected

        Returns:
            The count of cards (matching the condition if any)
        """
        return sum(count for _card, count in self.cards(condition))

    def _sorted_library(self) -> Generator[CardsBlock, None, None]:
        """A generator that yields library cards sorted by type and name.

        Yields:
            "type", [{card_name: count}]
        """

        def _type_index(card_count):
            return config.TYPE_ORDER.index(
                "/".join(sorted(vtes.VTES[card_count[0]]["Type"]))
            )

        library_cards = sorted(
            self.cards(vtes.VTES.is_library),
            key=lambda a: (
                _type_index(a),
                utils.normalize(vtes.VTES.vekn_name(vtes.VTES[a[0]])),
            ),
        )
        for kind, cards in itertools.groupby(library_cards, key=_type_index):
            # return a list so it can be iterated over multiple times
            yield kind, list(cards)

    def to_txt(self, wrap: int = 90) -> str:
        """A consistent deck display matching our parsing rules of TWDA.html

        Cards are displayed in the order given by the config.TYPE_ORDER list.
        The output is as closed as possible to the newest TWDA entries format.

        Args:
            wrap: If test should be wrapped (especially for comments )

        Returns:
            The normalized text version of the deck
        """
        lines = []
        if self.event:
            lines.append(self.event)
        if self.place:
            lines.append(self.place)
        if self.date:
            lines.append(arrow.get(self.date).format("MMMM Do YYYY"))
        if self.tournament_format:
            lines.append(self.tournament_format)
        if self.players_count:
            lines.append(f"{self.players_count} players")
        if self.player:
            lines.append(self.player)
        if self.event_link:
            lines.append(self.event_link)
        lines.append("")
        if self.score:
            lines.append(f"-- {self.score}")
            lines.append("")
        if self.name:
            lines.append(f"Deck Name: {self.name}")
        if self.author:
            lines.append(f"Created by: {self.author}")
        if self.comments:
            if self.name or self.author:
                lines.append("")
            if wrap and any(len(line) > wrap for line in self.comments.splitlines()):
                lines.extend(textwrap.wrap(self.comments, wrap))
            else:
                lines.append(self.comments)
        elif lines[-1] != "":
            lines.append("")
        cap = sorted(
            itertools.chain.from_iterable(
                [int(vtes.VTES[card]["Capacity"])] * count
                for card, count in self.cards(vtes.VTES.is_crypt)
            )
        )
        cap_min = sum(cap[:4])
        cap_max = sum(cap[-4:])
        cap_avg = sum(cap) / len(cap)
        lines.append(
            f"Crypt ({self.cards_count(vtes.VTES.is_crypt)} cards, "
            f"min={cap_min}, max={cap_max}, avg={round(cap_avg, 2):g})"
        )
        lines.append("-" * len(lines[-1]))
        max_name = (
            max(
                len(vtes.VTES.vekn_name(vtes.VTES[card]))
                for card, _count in self.cards(vtes.VTES.is_crypt)
            )
            + 1
        )
        max_disc = (
            max(
                len(" ".join(vtes.VTES[card]["Disciplines"]))
                for card, _count in self.cards(vtes.VTES.is_crypt)
            )
            + 1
        )
        max_title = (
            max(
                len(vtes.VTES[card].get("Title", ""))
                for card, _count in self.cards(vtes.VTES.is_crypt)
            )
            + 1
        )
        for card, count in self.cards(vtes.VTES.is_crypt):
            official_name = vtes.VTES.vekn_name(vtes.VTES[card])
            if official_name == "Camille Devereux, The Raven" and self.raven:
                lines.append(
                    f"{{}}x {{:<{max_name}}} {{:>2}} {{:<{max_disc}}} "
                    f"{{:<{max_title}}} {{}}:{{}}".format(
                        count - self.raven,
                        "Camille Devereux",
                        vtes.VTES[card]["Capacity"],
                        " ".join(sorted(vtes.VTES[card]["Disciplines"])),
                        vtes.VTES[card].get("Title", ""),
                        vtes.VTES[card]["Clan"][0],
                        vtes.VTES[card]["Group"],
                    )
                )
                official_name = "Raven"
                count = self.raven
            lines.append(
                f"{{}}x {{:<{max_name}}} {{:>2}} {{:<{max_disc}}} "
                f"{{:<{max_title}}} {{}}:{{}}".format(
                    count,
                    official_name,
                    vtes.VTES[card]["Capacity"],
                    " ".join(sorted(vtes.VTES[card]["Disciplines"])),
                    vtes.VTES[card].get("Title", ""),
                    vtes.VTES[card]["Clan"][0],
                    vtes.VTES[card]["Group"],
                )
            )
            if card in self.cards_comments:
                comment = self.cards_comments[card].replace("\n", " ").strip()
                lines[-1] += f" -- {comment}"
        lines.append(f"\nLibrary ({self.cards_count(vtes.VTES.is_library)} cards)")
        # form a section for each type with a header displaying the total
        for i, (kind, cards) in enumerate(self._sorted_library()):
            trifle_count = ""
            if kind == 0:
                trifle_count = sum(
                    count
                    for card, count in cards
                    if re.search(r"(t|T)rifle", vtes.VTES[card]["Card Text"])
                )
                if trifle_count:
                    trifle_count = f"; {trifle_count} trifle"
                else:
                    trifle_count = ""
            cr = "\n" if i > 0 else ""
            lines.append(
                f"{cr}{config.TYPE_ORDER[kind]} "
                f"({sum(count for card, count in cards)}{trifle_count})"
            )
            for card, count in cards:
                if card in self.cards_comments:
                    comment = self.cards_comments[card].replace("\n", " ").strip()
                    lines.append(
                        f"{count}x {vtes.VTES.vekn_name(vtes.VTES[card]):<23}"
                        f" -- {comment}"
                    )
                else:
                    lines.append(f"{count}x {vtes.VTES.vekn_name(vtes.VTES[card])}")
        return "\n".join(lines)

    def to_dict(self) -> dict:
        """A consistent deck serialization for the API

        Cards are listed in the order given by the config.TYPE_ORDER list.

        Returns:
            The serialized deck
        """
        ret = {
            "twda_id": self.id,
            "event": self.event,
            "place": self.place,
            "date": self.date,
            "tournament_format": self.tournament_format,
            "players_count": self.players_count,
            "player": self.player,
            "score": self.score,
            "name": self.name,
            "author": self.author,
            "comments": self.comments,
            "crypt": {
                "count": self.cards_count(vtes.VTES.is_crypt),
                "cards": [
                    {
                        "id": vtes.VTES[card]["Id"],
                        "count": count,
                        "name": card,
                        "comments": self.cards_comments.get(card),
                    }
                    for card, count in self.cards(vtes.VTES.is_crypt)
                ],
            },
            "library": {"count": self.cards_count(vtes.VTES.is_library), "cards": []},
        }

        # form a section for each type with a header displaying the total
        for kind, cards in self._sorted_library():
            ret["library"]["cards"].append(
                {
                    "type": config.TYPE_ORDER[kind],
                    "count": sum(count for card, count in cards),
                    "cards": [],
                }
            )
            for card, count in cards:
                ret["library"]["cards"][-1]["cards"].append(
                    {
                        "id": vtes.VTES[card]["Id"],
                        "count": count,
                        "name": card,
                        "comments": self.cards_comments.get(card),
                    }
                )
        return utils.json_clean(ret)



"""This module provides the `VTES` singleton: the VTES cards library.

If it has not been initialized, VTES will evaluate to False.
VTES must be configured with `VTES.configure()` before being used.
"""
from typing import Generator, List, Mapping, TextIO, Tuple, Union
import collections
import copy
import csv
import io
import itertools
import os
import pickle
import pkg_resources
import re
import requests
import tempfile
import typing
import warnings
import zipfile

import unidecode
import yaml

from . import config
from . import logging
from . import utils


StringGenerator = Generator[str, None, None]

logger = logging.logger


class _RulingsLinks(dict):
    def get_links(self, text: str) -> Generator[Tuple[str, str], None, None]:
        """Yield ruling references and links from a dict of links and a ruling text.

        Args:
            text: A ruling with one or more ruling references like [ZZZ 20000101].

        Yields:
            (reference, url) of the rulings contained in the text.
        """
        references = re.findall(r"\[[a-zA-Z]+\s[0-9-]+\]", text)
        if not references:
            warnings.warn(f"no reference in ruling: {text}")
        for reference in references:
            reference = reference[1:-1]
            yield reference, self[reference]


class _VTES:
    """VTES cards database that matches incomplete or misspelled card names.

    Keys are lower case card names and card Ids.
    Contains both crypt and library cards in a single dict.
    """

    def __init__(self):
        """Initialize the cards map

        `original_cards` and `rulings` are filled when loading the instance
        (either by Pickling or using self.load_from_vekn()).
        Other attributes are set on self.configure() call.


        Attributes:
            original_cards (dict): dict of original {id: card}
            rulings (dict): dict of {
                id: {
                    "Rulings": [ruling],
                    "Rulings Links": {
                        "Reference": "NNN YYYMMDD"
                        "URL": "http://www.example.com"
                    }
                }
            }
            cards (dict): dict of {card_name: card}
            completion (utils.Trie): used for card name completion
            search (dict): used for structured card search (cf. web API)
            fuzzy_threshold (int): do not try to fuzzy match text shorter than this
            safe_variants (bool): do not use card name variants shorter than the
                                  fuzzy threshold (avoids errors when parsing the TWDA)
        """
        self.original_cards = {}
        self.translations = {}
        self.rulings = {}
        self.cards = utils.FuzzyDict(aliases=config.REMAP)
        self.completion = utils.Trie()
        self.completion_i18n = collections.defaultdict(utils.Trie)
        self.search = {}
        self.search_i18n = collections.defaultdict(utils.Trie)  # only card text
        self.fuzzy_threshold = 6
        self.safe_variants = True

    def __getstate__(self) -> dict:
        """For pickle serialization."""
        return {
            "cards": self.original_cards,
            "translations": self.translations,
            "rulings": self.rulings,
        }

    def __setstate__(self, state: typing.Mapping) -> None:
        """For pickle deserialization."""
        self.original_cards = state.get("cards") or {}
        self.translations = state.get("translations") or {}
        self.rulings = state.get("rulings") or {}

    def save(self):
        pickle.dump(self, open(config.VTES_FILE, "wb"))

    def __reduce__(self) -> tuple:
        """For pickle serialization."""
        return (_VTES, (), self.__getstate__())

    def __bool__(self) -> bool:
        """Test for emptyness"""
        return bool(self.original_cards)

    def __getitem__(self, key):
        return self.cards[key]

    def __contains__(self, key):
        return key in self.cards

    def __setitem__(self, key, value):
        self.cards[key] = value

    def __delitem__(self, key):
        del self.cards[key]

    def items(self):
        return self.cards.items()

    def keys(self):
        return self.cards.keys()

    def values(self):
        return self.cards.values()

    def _yaml_get_card(self, text: str) -> dict:
        """Get a card from a YAML card reference (id|Name)

        This is used before self.configure() is called: self.cards is empty,
        self.original_cards is used.

        Args:
            text: a YAML card refreence of the form "id|Name"
        """
        card_id, card_name = text.split("|")
        card_id = int(card_id)
        card = self.original_cards[card_id]
        if self.get_name(card) != card_name:
            warnings.warn(
                f"rulings: {card_id} listed as {card_name} "
                f"instead of {self.get_name(card)}"
            )
        return card

    def load_from_vekn(self, save: bool = True) -> None:
        """Load the card database from vekn.net, with rulings

        Args:
            save: If True, card list is pickled to be retrieved faster later on.
        """
        self.original_cards.clear()
        self.translations.clear()
        self.rulings.clear()
        r = requests.request("GET", config.VEKN_VTES_URL)
        r.raise_for_status()
        with tempfile.NamedTemporaryFile("wb", suffix=".zip") as f:
            f.write(r.content)
            f.flush()
            z = zipfile.ZipFile(f.name)
            with z.open(config.VEKN_VTES_LIBRARY_FILENAME) as c:
                self.load_csv(io.TextIOWrapper(c, encoding=config.VEKN_VTES_ENCODING))
            with z.open(config.VEKN_VTES_CRYPT_FILENAME) as c:
                self.load_csv(io.TextIOWrapper(c, encoding=config.VEKN_VTES_ENCODING))
        # load translations:
        for lang, url in config.VEKN_VTES_I18N_URLS.items():
            self.translations[lang] = {}
            r = requests.request("GET", url)
            r.raise_for_status()
            with tempfile.NamedTemporaryFile("wb", suffix=".zip") as f:
                f.write(r.content)
                f.flush()
                z = zipfile.ZipFile(f.name)
                with z.open(config.VEKN_VTES_I18N_LIBRARY_FILENAME % lang) as c:
                    self.load_i18n_csv(
                        lang, io.TextIOWrapper(c, encoding=config.VEKN_VTES_ENCODING)
                    )
                with z.open(config.VEKN_VTES_I18N_CRYPT_FILENAME % lang) as c:
                    self.load_i18n_csv(
                        lang, io.TextIOWrapper(c, encoding=config.VEKN_VTES_ENCODING)
                    )
        # load rulings from local yaml files
        links = _RulingsLinks(
            yaml.safe_load(
                pkg_resources.resource_string("rulings", "rulings-links.yaml")
            )
        )
        cards_rulings = yaml.safe_load(
            pkg_resources.resource_string("rulings", "cards-rulings.yaml")
        )
        general_rulings = yaml.safe_load(
            pkg_resources.resource_string("rulings", "general-rulings.yaml")
        )
        for card, rulings in cards_rulings.items():
            card_name = self.get_name(self._yaml_get_card(card))
            self.rulings.setdefault(card_name, {"Rulings": [], "Rulings Links": []})
            self.rulings[card_name]["Rulings"].extend(rulings)
            self.rulings[card_name]["Rulings Links"].extend(
                {"Reference": reference, "URL": link}
                for ruling in rulings
                for reference, link in links.get_links(ruling)
            )
        for ruling in general_rulings:
            for card in ruling["cards"]:
                card_name = self.get_name(self._yaml_get_card(card))
                self.rulings.setdefault(card_name, {"Rulings": [], "Rulings Links": []})
                self.rulings[card_name]["Rulings"].append(ruling["ruling"])
                self.rulings[card_name]["Rulings Links"].extend(
                    {"Reference": reference, "URL": link}
                    for reference, link in links.get_links(ruling["ruling"])
                )
        if save:
            self.save()

    def configure(self, fuzzy_threshold: int = 6, safe_variants: bool = True) -> None:
        """Add alternative names for cards, set up completion

        Args:
            fuzzy_threshold: min text length to try fuzzy matching on.
            safe_variants: if False, more card name variants are used
                           (unsafe for TWDA parsing)
            translations: if True, add translations in search and completion
        """
        self.cards.clear()
        self.completion.clear()
        self.cards.threshold = fuzzy_threshold
        self.safe_variants = safe_variants
        # fill self.cards dict with all card names variants (including card ID)
        for card_id, card in self.original_cards.items():
            self[card_id] = card
            for name in self._get_name_variants(card, safe_variants):
                self[name] = card
        # we difflib match on AKA but not on REMAP
        # some items on REMAP are too short and match too many things.
        for alternative, card in config.AKA.items():
            self[alternative] = self[card]
        # build the completion tries for card names
        for name, card in self.cards.items():
            if isinstance(name, int):
                for lang in config.SUPPORTED_LANGUAGES:
                    if name in self.translations[lang]:
                        translation = self.translations[lang][name]["Name"]
                        self.completion_i18n[lang].add(translation, name)
                        while "," in translation:
                            translation = translation.rsplit(",", 1)[0]
                            self.completion_i18n[lang].add(translation, name)
            else:
                self.completion.add(name, int(card["Id"]))
        # build the cards map for structured search
        # text is a completion trie, other dicts just contain a set of the cards IDs
        self.search = {
            "text": utils.Trie(),
            "type": collections.defaultdict(set),
            "clan": collections.defaultdict(set),
            "trait": collections.defaultdict(set),
            "group": collections.defaultdict(set),
            "capacity": collections.defaultdict(set),
            "discipline": collections.defaultdict(set),
            "votes": set(),
            "intercept": set(),
            "bleed": set(),
            "stealth": set(),
        }
        for card in self.original_cards.values():
            self._add_card_to_search(card)
        # manual fixups
        for name, mods in config.SEARCH_EXCEPTIONS.items():
            for mod in mods:
                prefix = mod[0]
                key_path = mod[1:].split(":")
                try:
                    s = self.search
                    for key in key_path:
                        s = s[key]
                    if prefix == "+":
                        s.add(int(self[name]["Id"]))
                    elif prefix == "-":
                        s.remove(int(self[name]["Id"]))
                    else:
                        warnings.warn(f"{name} has an invalid prefix {prefix}")
                except KeyError:
                    warnings.warn(
                        f"{name} marked as exception for {key_path}, but is not listed."
                    )

    def _add_card_to_search(
        self, card: typing.Mapping, translations: bool = False
    ) -> None:
        """Add a card to self.search

        The card["Id"] is added to all matching sets.
        What matches or not has a lot do to with players usage and are design decisions:
            - requiring a title is listed as matching the sect
            - removing stealth is considered as intercept
            - stealthed actions are not considered as stealth cards
            - making a block fail is considered stealth
            - MERGED text of an advanced is taken into account for sect and title
            - combo and anarch tri-discipline cards are all listed under "*"
            - Vision uses the "vin" trigtam to distinguish from "vis"ceratika
        Args:
            card: A card dict
        """
        card_id = int(card["Id"])
        self.search["text"].add(card["Card Text"], card_id)
        for lang, translation in self.translations.items():
            if card_id in translation:
                self.search_i18n[lang].add(translation[card_id]["Card Text"], card_id)
        for type_ in card["Type"]:
            self.search["type"][type_.lower()].add(card_id)
        if set(card["Type"]) & {"Vampire", "Imbued"}:
            self.search["type"]["crypt"].add(card_id)
            # sect, including for MERGED versions
            if re.search(r"^(\[MERGED\] )?Sabbat", card["Card Text"]):
                self.search["trait"]["sabbat"].add(card_id)
            if re.search(r"^(\[MERGED\] )?Camarilla", card["Card Text"]):
                self.search["trait"]["camarilla"].add(card_id)
            if re.search(r"^(\[MERGED\] )?Laibon", card["Card Text"]):
                self.search["trait"]["laibon"].add(card_id)
            if re.search(r"^(\[MERGED\] )?Anarch", card["Card Text"]):
                self.search["trait"]["anarch"].add(card_id)
            if re.search(r"^(\[MERGED\] )?Independent", card["Card Text"]):
                self.search["trait"]["independent"].add(card_id)
        else:
            self.search["type"]["library"].add(card_id)
        if card.get("Clan"):
            for clan in card.get("Clan", []):
                self.search["clan"][clan.lower()].add(card_id)
        else:
            self.search["clan"]["none"].add(card_id)
        # CSV-listed traits
        if card.get("Sect"):
            self.search["trait"][card["Sect"].lower()].add(card_id)
        if card.get("Title"):
            self.search["trait"][card["Title"].lower()].add(card_id)
        if card.get("Group"):
            self.search["group"][card["Group"].lower()].add(card_id)
            if card["Group"].lower() == "any":
                for i in range(1, 7):
                    self.search["group"][str(i)].add(card_id)
        if card.get("Capacity"):
            capacity = card["Capacity"].lower()
            if len(card["Capacity"]) > 2:
                capacity = "+1"
            self.search["capacity"][capacity].add(card_id)
        if card.get("Title"):
            self.search["votes"].add(card_id)
        if re.search(r"(\d|X)\s+(v|V)ote", card["Card Text"]):
            self.search["votes"].add(card_id)
        # bleed/stealth/intercept bonuses
        if re.search(r"\+(\d|X)\s+(b|B)leed", card["Card Text"]):
            self.search["bleed"].add(card_id)
        if re.search(r"\+(\d|X)\s+(i|I)ntercept", card["Card Text"]):
            self.search["intercept"].add(card_id)
        # do not include stealthed actions as stealth cards
        if re.search(
            r"\+(\d|X)\s+(s|S)tealth (?!Ⓓ|action|political)", card["Card Text"]
        ):
            self.search["stealth"].add(card_id)
        # stealth/intercept maluses count has bonus of the other (for library cards)
        # also list block denials as stealth
        if not set(card["Type"]) & {"Master", "Vampire", "Imbued"}:
            if re.search(r"-(\d|X)\s+(s|S)tealth", card["Card Text"]):
                self.search["intercept"].add(card_id)
            if re.search(r"-(\d|X)\s+(i|I)ntercept", card["Card Text"]):
                self.search["stealth"].add(card_id)
        if re.search(r"stealth .* to 0", card["Card Text"]):
            self.search["intercept"].add(card_id)
        if re.search(r"attempt fails", card["Card Text"]):
            self.search["stealth"].add(card_id)
        for discipline in card.get("Discipline", []):
            self.search["discipline"][discipline.lower()].add(card_id)
        # add vampires with superior disciplines to both inf and sup sets
        for discipline in card.get("Disciplines", []):
            # Visceratika and Vision have the same trigram, discriminate
            if discipline.lower() == "vis" and "Imbued" in card["Type"]:
                discipline = "vin"
            self.search["discipline"][config.DIS_MAP[discipline.lower()]].add(card_id)
            if discipline in config.DIS_MAP:
                self.search["discipline"][config.DIS_MAP[discipline]].add(card_id)
        if not card.get("Disciplines") and not card.get("Discipline"):
            self.search["discipline"]["none"].add(card_id)
        # "Discipline" singular is a library-only key
        if len(card.get("Discipline", [])) > 1:
            self.search["discipline"]["*"].add(card_id)
        # FLIGHT is not listed as a discipline, treat it as such anyway
        if "[FLIGHT]" in card["Card Text"]:
            self.search["discipline"]["flight"].add(card_id)
        # city trait
        city = re.search(
            r"(?:Prince|Baron|Archbishop) of ((?:[A-Z][A-Za-z\s-]{3,})+)",
            card["Card Text"],
        )
        if city:
            city = city.group(1).lower()
            if city[-5:] == "as a ":
                city = city[:-5]
            if city == "washington":
                city = "washington, d.c."
            self.search["trait"][city].add(card_id)
            self.search["votes"].add(card_id)
        # add generic title requirements to all titles of the sect
        if re.search(r"Requires a( ready)? titled (S|s)abbat", card["Card Text"]):
            self.search["trait"]["bishop"].add(card_id)
            self.search["trait"]["archbishop"].add(card_id)
            self.search["trait"]["cardinal"].add(card_id)
            self.search["trait"]["regent"].add(card_id)
            self.search["trait"]["priscus"].add(card_id)
            self.search["trait"]["sabbat"].add(card_id)
        if re.search(r"Requires a( ready)? titled (C|c)amarilla", card["Card Text"]):
            self.search["trait"]["primogen"].add(card_id)
            self.search["trait"]["prince"].add(card_id)
            self.search["trait"]["justicar"].add(card_id)
            self.search["trait"]["inner circle"].add(card_id)
            self.search["trait"]["imperator"].add(card_id)
            self.search["trait"]["camarilla"].add(card_id)
        # this should match unnamed independant titles too, but we have no trait for it
        if re.search(r"Requires a( ready)? titled vampire", card["Card Text"]):
            self.search["trait"]["primogen"].add(card_id)
            self.search["trait"]["prince"].add(card_id)
            self.search["trait"]["justicar"].add(card_id)
            self.search["trait"]["inner circle"].add(card_id)
            self.search["trait"]["imperator"].add(card_id)
            self.search["trait"]["bishop"].add(card_id)
            self.search["trait"]["archbishop"].add(card_id)
            self.search["trait"]["cardinal"].add(card_id)
            self.search["trait"]["regent"].add(card_id)
            self.search["trait"]["priscus"].add(card_id)
            self.search["trait"]["baron"].add(card_id)
            self.search["trait"]["liaison"].add(card_id)
            self.search["trait"]["magaji"].add(card_id)
            self.search["trait"]["kholo"].add(card_id)
        if re.search(r"Requires a( ready)? (M|m)agaji", card["Card Text"]):
            self.search["trait"]["magaji"].add(card_id)
            self.search["trait"]["laibon"].add(card_id)
        # consider sects as traits,
        # but do not match them just anywhere in the card text
        if re.search(r"Requires a( ready)? (S|s)abbat", card["Card Text"]):
            self.search["trait"]["sabbat"].add(card_id)
        if re.search(r"Requires a( ready)? (C|c)amarilla", card["Card Text"]):
            self.search["trait"]["camarilla"].add(card_id)
        if re.search(r"Requires a( ready)? (L|l)aibon", card["Card Text"]):
            self.search["trait"]["laibon"].add(card_id)
        if re.search(
            r"Requires a( ready|n)( (I|i)ndependent or)? (A|a)narch",
            card["Card Text"],
        ):
            self.search["trait"]["anarch"].add(card_id)
        if re.search(r"Requires an (I|i)ndependent", card["Card Text"]):
            self.search["trait"]["independent"].add(card_id)
        # match other traits anywhere in the text
        # add specific title requirements to the matching sect
        for trait in re.findall(
            f"({'|'.join(config.TRAITS)})", card["Card Text"].lower()
        ):
            self.search["trait"][trait].add(card_id)
            if trait in {"primogen", "prince", "justicar", "inner circle", "imperator"}:
                self.search["trait"]["camarilla"].add(card_id)
            if trait in {"bishop", "archbishop", "cardinal", "regent", "priscus"}:
                self.search["trait"]["sabbat"].add(card_id)
            if trait in {"baron"}:
                self.search["trait"]["anarch"].add(card_id)
            if trait in {"magaji", "kholo"}:
                self.search["trait"]["laibon"].add(card_id)

    def complete(self, partial_name: str) -> List:
        """Card name completion.

        It uses a Trie to match trigrams.
        Matches on the start of the name are returned first,
        other matches are returned alphabetically.

        Args:
            partial_name: Parts of the name (can contain spaces)

        Returns:
            A sorted list of results, from most likely to less likely
        """
        ret = [
            [self.get_name(card_id), score]
            for card_id, score in self.completion.search(partial_name).items()
        ]
        return [x[0] for x in sorted(ret, key=lambda x: (-x[1], x[0]))]

    def complete_i18n(self, partial_name: str, lang: str) -> List:
        """Card name completion, with translations.

        Searches for both english and given language completions.
        The maximum score is used for ordering (then, alphabetically).

        Args:
            partial_name: Parts of the name (can contain spaces)

        Returns:
            A list of dicts where the keys are language codes
            and the values are the card names in those languages.
        """
        search = self.completion.search(partial_name)
        if lang in self.completion_i18n:
            search |= self.completion_i18n[lang].search(partial_name)
        ret = []
        for card_id, score in search.items():
            ret.append([score, {"en": self.get_name(card_id)}])
            if card_id in self.translations.get(lang, {}):
                ret[-1][1].update({lang: self.translations[lang][card_id]["Name"]})
        return [
            x[1]
            for x in sorted(ret, key=lambda x: (-x[0], x[1].get(lang) or x[1]["en"]))
        ]

    def trait_choices(self, trait: str) -> set:
        """Get all registered values for a given card trait.

        Available traits:
        - Adv
        - Aka
        - Artist
        - Banned
        - Blood Cost
        - Burn Option
        - Capacity
        - Card Text
        - Clan
        - Conviction Cost
        - Discipline (Library card)
        - Disciplines (Crypt card)
        - Draft
        - Flavor Text
        - Group
        - Id
        - Name
        - Pool Cost
        - Requirement
        - Set
        - Title
        - Type

        Empty values are removed.

        Args:
            trait: The trait to get choices for.

        Returns:
            All possible values
        """
        return set(
            itertools.chain.from_iterable(
                card[trait] for card in self.values() if card.get(trait)
            )
        )

    @property
    def disciplines(self) -> list:
        """List of disciplines"""
        return sorted(self.trait_choices("Discipline"))

    @property
    def types(self) -> list:
        """List of card types"""
        return sorted(self.trait_choices("Type"))

    @property
    def clans(self) -> list:
        """List of clans"""
        return sorted(self.trait_choices("Clan"))

    def get_name(self, card: Union[str, int, Mapping]) -> str:
        """Returns official name for a card or card name."""
        if not isinstance(card, collections.abc.Mapping):
            card = self[card]
        name = card["Name"]
        if name[-5:] == ", The":
            name = "The " + name[:-5]
        name = name.replace("(TM)", "™")
        if name[-6:] != " (ADV)" and card.get("Adv"):
            name += " (ADV)"
        return name

    def vekn_name(self, card: Union[str, Mapping]) -> str:
        """Returns the VEKN name for a card (suffix The, (TM) instead of ™)."""
        if not isinstance(card, collections.abc.Mapping):
            card = self[card]
        name = card["Name"]
        if name[:4] == "The ":
            name = name[4:] + ", The"
        if name[-6:] != " (ADV)" and card.get("Adv"):
            name = name + " (ADV)"
        name = name.replace("™", "(TM)")
        return name

    def normalized(self, card: Union[str, Mapping]) -> Mapping:
        """Returns a JSON-serializable version of the card.

        It makes sure the name is the "standard" one (with (ADV) for advanced vampires)
        and adds a field with the card image URL and rulings.
        """
        if not isinstance(card, collections.abc.Mapping):
            card = self[card]
        data = copy.copy(card)
        name = self.get_name(card)
        data["Name"] = name
        file_name = utils.normalize(name)
        file_name = file_name[4:] + "the" if file_name[:4] == "the " else file_name
        file_name, _ = re.subn(r"""\s|,|\.|-|—|'|:|\(|\)|"|!""", "", file_name)
        data["Image"] = f"{config.IMAGES_ROOT}{file_name}.jpg"
        data.update(self.rulings.get(name, {}))
        for lang, translations in self.translations.items():
            if int(card["Id"]) in translations:
                data.setdefault("Translations", {})
                data["Translations"][lang] = translations[int(card["Id"])]
        return data

    def _get_name_variants(self, card: Mapping, safe: bool = True) -> StringGenerator:
        """Yields all name variants for a card

        Add the " (ADV)" suffix for advanced vampires
        Yield an ASCII variant for non-ASCII names
        Yield a straightforward name for deported article:
            "Second Tradition, The" -> "The Second Tradition"
        Yield a shorten version for comma names:
            "Akhenaten, The Sun Pharaoh" -> "Akhenaten"
        Apply all of the above for official "Aka" variants

        Important examples are in test.

        Args:
            card: the card
            safe: if False, yield more variants (unsafe for TWDA parsing)

        Yields:
            All name variants
        """
        advanced_suffix = " (ADV)" if card.get("Adv") else ""

        def name_variants(name):
            name = name.strip()
            if not name:
                return
            yield name + advanced_suffix
            if "(TM)" in name:
                yield name.replace("(TM)", "™") + advanced_suffix
            ascii_variant = unidecode.unidecode(name)
            if ascii_variant != name:
                yield ascii_variant + advanced_suffix
            if name[-5:] == ", The":
                yield from name_variants("The " + name[:-5])
            # suffix removal
            # --------------
            # Note the following code also removes the ", The" particle.
            # Calling it multiple times can get all particles removed,
            # eg. "Rumor Mill, Tabloid Newspaper, The" will yield:
            # - "Rumor Mill, Tabloid Newspaper"
            # - "Rumor Mill"
            #
            # ':' can not be seen as a suffix marker,
            # because we would miss with "Praxis Seizure:" and "Crusade:" cards.
            # Too bad, because we would like to get the "Second tradition:" variants.
            # Henceforth, these variants are handled manually in config.REMAP.
            if "," in name:
                alternative = name.rsplit(",", 1)[0]
                # We do not shorten too much or anything will fuzzy match
                # eg. "Line" (without ", The") could mismatch "Redline" or "Zip Line".
                # Still, "Gunther" should match "Gunther, Beast Lord'
                if not safe or len(alternative) > 6:
                    yield from name_variants(alternative)

        yield from name_variants(card["Name"])
        # multiple aliases are separated by semicolumn
        for alias in card.get("Aka", "").split(";"):
            yield from name_variants(alias)

    def load_i18n_csv(self, lang, stream: TextIO) -> None:
        """Load a crypt or library Translation CSV (VEKN format)"""
        for card in csv.DictReader(stream):
            self.translations[lang][int(card["Id"])] = {
                "Name": card["Name " + lang],
                "Card Text": card["Card Text"],
                "Flavor Text": card.get("Flavor Text") or "",
            }

    def load_csv(self, stream: TextIO) -> None:
        """Load a crypt or library CSV (VEKN format)"""
        for card in csv.DictReader(stream):
            if "Set" in card:
                card["Set"] = [s.strip() for s in card["Set"].split(",") if s.strip()]
            if "Clan" in card:
                card["Clan"] = [c.strip() for c in card["Clan"].split("/") if c.strip()]
            if "Type" in card:
                card["Type"] = [t.strip() for t in card["Type"].split("/") if t.strip()]
            if "Card Text" in card:
                card["Card Text"] = card["Card Text"].replace("{", "")
                card["Card Text"] = card["Card Text"].replace("}", "")
                card["Card Text"] = card["Card Text"].replace("(D)", "Ⓓ ")
            if "Disciplines" in card:
                card["Disciplines"] = [
                    d.strip() for d in card["Disciplines"].split(" ") if d.strip()
                ]
            if "Requirement" in card:
                card["Requirement"] = [
                    r.strip() for r in card["Requirement"].split(",") if r.strip()
                ]
            if "Discipline" in card:
                card["Discipline"] = [
                    d.strip()
                    for p in card["Discipline"].split("/")
                    for d in p.split("&")
                    if d.strip()
                ]
            if "Burn Option" in card:
                card["Burn Option"] = bool(card["Burn Option"])
            if "Adv" in card:
                card["Adv"] = bool(card["Adv"])
            # remove deprecated discipline columns for crypt cards
            for disc in config.DISC_AKA.values():
                if disc in card:
                    del card[disc]
            card_id = int(card["Id"])
            if card_id not in self.original_cards:
                self.original_cards[card_id] = {}
            self.original_cards[int(card["Id"])].update(card)

    def is_crypt(self, card_name: str) -> bool:
        """A function to check if a card is a crypt card"""
        return set(self[card_name]["Type"]).issubset({"Vampire", "Imbued"})

    def is_library(self, card_name: str) -> bool:
        """A function to check if a card is a library card"""
        return not self.is_crypt(card_name)

    def is_disc(self, card_name: str, discipline: str) -> bool:
        """A function to check if a card is of given discipline

        Handles multi-valued cards too (combo or outferior option).
        """
        return utils.normalize(discipline) in {
            utils.normalize(d) for d in self[card_name].get("Discipline", [])
        }

    def no_disc(self, card_name: str) -> bool:
        """A function to check if a card requires no discipline"""
        return not self[card_name].get("Discipline", [])

    def is_type(self, card_name: str, type_: str) -> bool:
        """A function to check if a card is of the given type

        Handles multiple types too (e.g. Modifier/Reaction)
        """
        return utils.normalize(type_) in {
            utils.normalize(t) for t in self[card_name]["Type"]
        }

    def is_clan(self, card_name: str, clan: str) -> bool:
        """A function to check if a card is of the given clan (or requires said clan)

        Handles multiple clans too (e.g. Gangrel/Gangrel Antitribu)
        """
        return utils.normalize(clan) in {
            utils.normalize(c) for c in self[card_name]["Clan"]
        }

    def str_to_card_and_count(self, string: str) -> dict:
        string = string.lower()


try:
    if not os.path.exists(config.VTES_FILE) or os.stat(config.VTES_FILE).st_size == 0:
        raise FileNotFoundError(config.VTES_FILE)
    VTES = pickle.load(open(config.VTES_FILE, "rb"))
except (FileNotFoundError, EOFError):
    VTES = _VTES()  # evaluates to False as it is empty



"""Generic utilities used by the library.
"""
from typing import Any, Dict, Hashable, List, Sequence
import argparse
import arrow
import collections
import datetime
import difflib
import re

import unidecode

from . import logging

logger = logging.logger


def normalize(s: Any):
    """Normalize a string for indexing: unidecode and lowercase."""
    if not isinstance(s, str):
        return s
    return unidecode.unidecode(s).lower().strip()


class LRU(collections.OrderedDict):
    """Python recipy for LRU cache."""

    def __init__(self, maxsize: int = 256, *args, **kwargs):
        self.maxsize = maxsize
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):
        value = super().__getitem__(key)
        self.move_to_end(key)
        return value

    def __setitem__(self, key, value):
        if key in self:
            self.move_to_end(key)
        super().__setitem__(key, value)
        if len(self) > self.maxsize:
            oldest = next(iter(self))
            del self[oldest]


class FuzzyDict(dict):
    """A dict providing "fuzzy matching" of its keys.

    It matches keys that are "close enough" if there are no exact match,
    and provides athe ability to specify aliases for certain keys.

    Aliases are only matched exactly (not closely like normal keys).

    Attributes:
        threshold: minimum string length for fuzzy matching
        cutoff: minimum similarity to consider a close candidate an actual match
        aliases: {alias_key: actual_key} adding not-so-closes aliases
    """

    def __init__(
        self,
        threshold: int = 6,
        cutoff: float = 0.85,
        aliases: dict = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        # cache
        self.cache = LRU()
        self.keys_cache = None
        #: only use fuzzy match if len(key) >= to THRESHOLD
        self.threshold = threshold
        #: Similarity cutoff for underlying difflib.get_close_matches()
        self.cutoff = cutoff
        #: When a key is not in dict, use an alias if it exists
        self.aliases = aliases

    def _fuzzy_match(self, key: Hashable) -> Any:
        """Use difflib to match incomplete or misspelled keys

        It uses a cache to accelerate matching of common errors.

        Args:
            name: the card name

        Returns:
            The matched card name, or None. This method does not raise
        """
        if key in self.cache:
            return self.cache[key]
        if not isinstance(key, collections.abc.Sequence):
            return None
        if len(key) < self.threshold:
            return None
        result = difflib.get_close_matches(
            key, self.sequence_keys(), n=1, cutoff=self.cutoff
        )
        if result:
            result = result[0]
            logger.info('"{}" matched "{}"', key, result)
            self.cache[key] = result
            return result
        return None

    def sequence_keys(self) -> List[Sequence]:
        """Return all keys that are sequences and can be fuzzy matched."""
        if not self.keys_cache:
            self.keys_cache = [
                k for k in self.keys() if isinstance(k, collections.abc.Sequence)
            ]
        return self.keys_cache

    def __getitem__(self, key: Hashable):
        """Get a key, try to find a good matching.

        It uses lowercase only, plus config.REMAP for common errors and abbreviations.
        It also uses difflib to fuzzy match incomplete or misspelled names

        Args:
            key: the key to search for
        """
        key = normalize(key)
        try:
            return super().__getitem__(key)
        except KeyError:
            # aliases must match exactly
            if key in self.aliases:
                return super().__getitem__(normalize(self.aliases[key]))
            fuzzy_match = self._fuzzy_match(key)
            if fuzzy_match:
                return super().__getitem__(fuzzy_match)
            raise

    def __contains__(self, key):
        try:
            self.__getitem__(key)
            return True
        except KeyError:
            return False

    def __setitem__(self, key, value):
        self._clear_cache()
        return super().__setitem__(normalize(key), value)

    def __delitem__(self, key):
        self._clear_cache()
        return super().__delitem__(normalize(key))

    def _clear_cache(self):
        self.cache.clear()
        self.keys_cache = None


class Trie(collections.defaultdict):
    """A Trie structure for text search.

    It relies on Python dict and has the following structure:
    {prefix: {reference: score}}.

    When a (text, reference) couple is entered in the Trie,
    every prefix of every word of the text is added to the Trie,
    pointing to the reference with a score depending on the length
    of the prefix and the position in the text (matching forst word is worth double).

    The matches are case-insensitive and use unidecode to handle unicode characters.
    """

    def __init__(self):
        super().__init__(lambda: collections.defaultdict(int))

    @staticmethod
    def _split(text):
        return re.sub(r"[/:,\(\)'\"]", " ", normalize(text)).split()

    def add(self, text: str, reference: Any) -> None:
        """Add text to the Trie

        Args:
            text: The text to add.
            reference: The reference to return on a match
        """
        for e, part in enumerate(Trie._split(text)):
            for i in range(1, len(part) + 1):
                self[part[:i]][reference] += (
                    # double score for matching name start
                    i
                    * (2 if e == 0 else 1)
                )

    def search(self, text: str) -> Dict[str, int]:
        """Search text into the Trie

        The match is case-insensitive and use unidecode, but is otherwise exact.
        Matching on the first word of a key scores double.
        Only candidates matching all words are returned.

        Args:
            text: The text to search
        Returns:
            References ordered by score
        """
        ret = None
        for part in Trie._split(text):
            # a word can match multiple parts of a key to one reference
            # take the highest score
            matches = {}
            for reference, score in self.get(part, {}).items():
                matches[reference] = max(matches.get(reference, 0), score)
            # match all words of given text
            if ret is not None:
                ret = collections.Counter(
                    {k: v + matches[k] for k, v in ret.items() if k in matches}
                )
            else:
                ret = collections.Counter(matches)
        return ret


def json_clean(obj: Any) -> Any:
    """Remove empty values in depth, returns a JSON-serializable dict/list structure."""
    # Basic types
    if obj is None or isinstance(obj, (bool, int, float)):
        return obj
    if isinstance(obj, (str, bytes)):
        return obj if obj else None
    # Mappings
    if isinstance(obj, collections.abc.Mapping):
        ret = {}
        for key, value in obj.items():
            value = json_clean(value)
            if value is not None:
                ret[key] = value
        return ret if ret else None
    # Dates
    if isinstance(obj, arrow.Arrow):
        obj = obj.datetime
    if isinstance(obj, datetime.date):
        return json_clean(obj.isoformat())
    # Iterables (should be last)
    if isinstance(obj, collections.abc.Iterable):
        ret = []
        for value in obj:
            value = json_clean(value)
            if value is not None:
                ret.append(value)
        return ret if ret else None
    # Anything empty is nulled, then removed
    return obj if obj else None


class NargsChoiceWithAliases(argparse.Action):
    """Choices with nargs +/*: this is a known issue for argparse
    cf. https://bugs.python.org/issue9625
    """

    CHOICES = []
    ALIASES = {}
    CASE_SENSITIVE = False

    def __call__(self, parser, namespace, values, option_string=None):
        if not self.CASE_SENSITIVE:
            values = [v.lower() for v in values]
        if values:
            for value in values:
                if value not in self.CHOICES and value not in self.ALIASES:
                    raise argparse.ArgumentError(
                        self, f"invalid choice: {value} (choose from {self.CHOICES})"
                    )
        setattr(
            namespace,
            self.dest,
            [
                value if value in self.CHOICES else self.ALIASES[value]
                for value in values
            ],
        )



"""This module provides the `TWDA` singleton: Tournament Wining Decks Archive.

If it has not been initialized, TWDA will evaluate to False.
TWDA must be configured with `TWDA.configure()` before being used.
"""
from typing import TextIO
import collections
import io
import itertools
import os
import pkg_resources  # part of setuptools
import pickle
import re

import arrow
import requests

from . import config
from . import logging
from . import deck
from . import utils

logger = logging.logger


class _TWDA(collections.OrderedDict):
    """An OrderedDict of the TWDA. Parsing TWDA.html is the hard part.

    Attributes:
        spoilers (dict): cards played in over 25% of decks
        tail_re (str): regexp used to parse tail part of card line
    """

    def load_from_vekn(self, limit: int = None, save: bool = True) -> None:
        """Load from vekn.net

        Args:
            limit: Maximum number of decks to load (used to speed up tests)
        """
        r = requests.request("GET", config.VEKN_TWDA_URL)
        r.raise_for_status()
        self.load_html(io.StringIO(r.content.decode("utf-8")), limit, save)

    def load_html(self, source: TextIO, limit: int = None, save: bool = True) -> None:
        """Load from TWDA.html

        The TWDA is then pickled for future use, to avoid loading it too often.

        Args:
            source (file): The HTML
            limit: Maximum number of decks to load (used to speed up tests)
        """
        self.clear()
        id_, buffer, offset = "", None, 0
        for index, line in enumerate(source, 1):
            try:
                id_ = re.match(r"^<a id=([^\s]*)\s", line).group(1)
            except AttributeError:
                pass
            if re.match(r"^<hr><pre>$", line):
                buffer = io.StringIO()
                offset = index
            elif re.match(r"^</pre>", line):
                buffer.seek(0)
                # p = parser.Parser(id=id_)
                # self[id_] = p.deck
                if pkg_resources.resource_exists("twda_fix", f"{id_}.html"):
                    buffer = io.StringIO(
                        pkg_resources.resource_string("twda_fix", f"{id_}.html").decode(
                            "utf-8"
                        )
                    )
                self[id_] = deck.Deck.from_txt(buffer, id=id_, offset=offset, twda=True)
                # p.parse(buffer, offset=offset, twda=True)
                # self[id_].parse(buffer, offset=offset, twda=True)
                if limit and len(self) >= limit:
                    break
            elif buffer:
                buffer.write(line)
        logger.info("TWDA loaded")
        if save:
            pickle.dump(TWDA, open(config.TWDA_FILE, "wb"))

    def configure(
        self,
        date_from: arrow.Arrow = None,
        date_to: arrow.Arrow = None,
        min_players: int = 0,
        spoilers: bool = True,
    ) -> None:
        """Prepare the TWDA, taking date and players count filters into account.

        Args:
            date_from: filter out decks before this date
            date_to: filter out decks after this date
            min_players: if > 0, filter out decks with less players
            spoilers: if True, compute spoilers to filter them out for analysis
        """
        if date_from:
            for key in [
                key for key, value in self.items() if value.date < date_from.date()
            ]:
                del self[key]
        if date_to:
            for key in [
                key for key, value in self.items() if value.date >= date_to.date()
            ]:
                del self[key]
        if min_players:
            for key in [
                key
                for key, value in self.items()
                if value.players_count and value.players_count < min_players
            ]:
                del self[key]
        if spoilers and len(self) > 50:
            self.spoilers = {
                name: count / len(self)
                for name, count in collections.Counter(
                    itertools.chain.from_iterable(d.keys() for d in self.values())
                ).items()
                if count > len(self) / 4
            }
            logger.debug("Spoilers: {}", self.spoilers)
        else:
            self.spoilers = {}
        self.by_author = collections.defaultdict(list)
        for id, d in self.items():
            if d.author:
                self.by_author[utils.normalize(d.author)].append(id)
            if d.player:
                self.by_author[utils.normalize(d.author)].append(id)


try:
    if not os.path.exists(config.TWDA_FILE) or os.stat(config.TWDA_FILE).st_size == 0:
        raise FileNotFoundError(config.TWDA_FILE)
    TWDA = pickle.load(open(config.TWDA_FILE, "rb"))
except (FileNotFoundError, EOFError):
    TWDA = _TWDA()  # evaluates to False as it is empty
